
#ifndef BUFFER_H
#define BUFFER_H


#include <vector>
#include <cstring>

#include <string>
#include "../Atlas.h" // for common |using| declarations
#include "bitmap.h" // so that the type |BitMap| is complete
#include "sl_list.h" 

#include <fstream>



namespace atlas
{ namespace interpreter
  {
    
    class String_pool
    { public:
        static const size_t default_block_size=512;
        String_pool(size_t b=default_block_size); // create pool and set block size
        char* store(const char* s, size_t length); // store string of given length
        ~String_pool();
      private:
        String_pool(const String_pool&); // copying forbidden
        String_pool& operator=(const String_pool&); // assignment forbidden
    
        const size_t block_size; // block size of the pool
        char* start; // start of block in current usage
        char* point; // first character available
        size_t chars_left; // number of characters left in current block
        const String_pool* prev;
          // start of list of previous blocks, for deallocation
    };
    
    typedef unsigned short id_type; // type of value representing identifiers
    
    class Hash_table
    {
    public:
      static const id_type empty; // value (|~0u|) reserved for empty slots
    private: // data members
      String_pool pool;
      id_type mod;  // hash modulus, the number of slots present
      std::vector<id_type> hash_tab;
      std::vector<const char*> name_tab;
    
    public: // interface
      Hash_table(size_t init=initial_hash_mod
                ,size_t block_size=String_pool::default_block_size);
    
      id_type nr_entries() const { return name_tab.size(); }
      const char* name_of(id_type nr) const { return name_tab[nr]; }
      bool knows(const char* name) const;
    
      id_type match(const char* s, size_t l) { return do_match(s,l,true); }
      id_type match_literal(const char* s) /* literal null-terminated string */
        { return do_match(s,std::strlen(s),false); }
    private: // auxiliary functions
      static const id_type initial_hash_mod=97;
      id_type hash(const char* s, size_t l) const; // auxiliary hash function
      id_type do_match(const char* s, size_t l, bool copy_string);
      size_t max_fill() const { return static_cast<id_type>(mod*3L/4); }
                      // maximum of entries in use
    };
    
    #include <iosfwd>
    
    class BufferedInput
    { typedef char* (*rl_type)(const char* );
         // |typedef| avoids inclusion of \.{readline} headers
      typedef void (*add_hist_type) (const char* );  // this one too
      
      struct input_record
      { std::ifstream f_stream; // the actual stream record
        id_type name; // identifies the file name for |stream|
        unsigned long line_no; // this refers to the older input stream!
      
        input_record(BufferedInput&, const char* file_name);
        input_record(const input_record& rec) = delete; // cannot copy
        input_record& operator=(const input_record& rec) = delete; // or assign
      };
     // a local class
    
    
      
      std::istream& base_stream;
      std::string line_buffer; // current line
      const char* p; // buffer pointer
      const char *const prompt, *const prompt2; // primary and secondary prompt
      const char* def_ext; // default file extension
      std::string temp_prompt; // varying addition to secondary prompt
      const rl_type readline; // readline function
      const add_hist_type add_hist; // history function
      unsigned long line_no; // current line number
      int cur_lines,prompt_length; // local variables
      containers::mirrored_sl_list<input_record> input_stack;
       // active input streams
      Hash_table input_files_seen; // files successfully opened at least once
      BitMap input_files_completed; // marks those files that were read without error
      std::istream* stream; 
    
    
     public:
        BufferedInput (std::istream& s);
            // associate line buffer to raw input stream
        BufferedInput (const char* prompt
                      , rl_type rl=nullptr, add_hist_type=nullptr 
    		  , const char* prompt2="> "
                      , const char* def_ext=".at");
            // use |stdin|, maybe with readline
    
        char shift (); // inspect a new character
        void unshift (); // back up so last character will be reconsidered
        bool eol () const; // end of line: |true| if |shift| would return a newline
        bool getline ();
             // fetch a new line to replace current one; |true| if successful
        bool push_file (const char* file_name,bool skip_seen);
          // |true| if successful
        void close_includes();
        
        id_type current_file() const
        { return input_stack.empty() ? Hash_table::empty : input_stack.back().name; }
        const char* name_of(id_type f) const
        { return
           f==Hash_table::empty ? "<standard input>" : input_files_seen.name_of(f); }
        const char* cur_fname() const
          { return name_of(current_file()); }
        
        unsigned int include_depth() const { return input_stack.size(); }
        const char* point() const  {return p==nullptr ? line_buffer.c_str(): p;}
        void set_line_no (unsigned long l) 
        { line_no=l; }
        void locate (const char* p, int& line, int& column) const;
        
        void show_range
         (std::ostream& out,unsigned long l0, int c0, unsigned long l1, int c1)
         const;
        
        void push_prompt(char c);
        char top_prompt() const;
        void pop_prompt();
        void reset();
    
    private:
        void pop_file();
    };

    
    inline bool BufferedInput::eol() const { return p==nullptr or *p=='\0'; }
    
    inline char BufferedInput::shift()
    { if (eol())
        if (not getline())
          { p=nullptr; return '\0'; } // signals file end
      return *p++;
    }
    
    inline void BufferedInput::unshift()
    { if (p!=nullptr and p>line_buffer.data())
          --p;
    }

    
    extern BufferedInput* main_input_buffer;

    
    unsigned int input_path_size();
    const std::string& input_path_component(unsigned int i);

  }
}
#endif

